/** DirectoryPage **/
#breadcrumbs>a:hover {
    text-decoration: none;
}

#breadcrumbs>a .dir {
    padding: 0 0.25em;
}

#breadcrumbs>a:first-child:hover::before,
#breadcrumbs>a .dir:hover {
    text-decoration: underline;
}

#breadcrumbs>a:first-child::before {
    content: "🏠";
}

#breadcrumbs>a:last-child {
    color: #ff0d68;
}

main a {
    color: inherit;
    font-weight: bold;
}

table.autoindex {
    width: 100%;
    font-family: monospace;
    font-size: 1.25rem;
}

table.autoindex tr {
    display: flex;
}

table.autoindex tr:hover {
    background-color: #ddd;
}

table.autoindex td {
    margin: 0 0.5rem;
}

table.autoindex td:first-child {
    flex: 1;
}

table.autoindex td:nth-child(2) {
    text-align: right;
}

table.autoindex td:last-child {
    text-align: right;
}


@media (prefers-color-scheme: dark) {
    table.autoindex tr:hover {
        background-color: #222;
    }
}
